package com.fjhx.service.impl;

import cn.hutool.core.bean.BeanUtil;
import cn.hutool.core.io.FileUtil;
import cn.hutool.core.lang.Assert;
import cn.hutool.core.thread.ThreadUtil;
import com.alibaba.fastjson.JSON;
import com.alibaba.fastjson.JSONObject;
import com.baomidou.mybatisplus.core.toolkit.IdWorker;
import com.baomidou.mybatisplus.core.toolkit.ObjectUtils;
import com.baomidou.mybatisplus.core.toolkit.Wrappers;
import com.fjhx.config.TaskPoolConfig;
import com.fjhx.config.redis.RedisCache;
import com.fjhx.constants.RedisConstant;
import com.fjhx.constants.SendConstant;
import com.fjhx.entity.*;
import com.fjhx.enums.SendEventEnum;
import com.fjhx.service.*;
import com.fjhx.utils.EmailEngineUtil;
import com.fjhx.utils.RetryUtil;
import com.fjhx.vo.*;
import lombok.extern.slf4j.Slf4j;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.beans.factory.annotation.Qualifier;
import org.springframework.stereotype.Service;
import org.springframework.transaction.annotation.Transactional;

import java.util.ArrayList;
import java.util.List;
import java.util.concurrent.Executor;
import java.util.stream.Collectors;

@Slf4j
@Service
public class AccountServiceImpl implements IAccountService {

    @Autowired
    private IEmailInfoService emailInfoService;

    @Autowired
    private IEmailMailboxService emailMailboxService;

    @Autowired
    private IEmailMessageService emailMessageService;

    @Autowired
    private IEmailMessageSendService emailMessageSendService;

    @Autowired
    private IEmailMessageAttachmentService emailMessageAttachmentService;

    @Qualifier(TaskPoolConfig.taskExecutor)
    @Autowired
    private Executor executor;

    @Transactional(rollbackFor = Exception.class)
    @Override
    public EmailInfo binding(BindingVo bindingVo) {

        String email = bindingVo.getEmail();

        EmailInfo emailInfo = emailInfoService.getOne(Wrappers.<EmailInfo>lambdaQuery().eq(EmailInfo::getEmail, email));
        // 如果存在，直接返回邮箱信息
        if (emailInfo != null) {
            return emailInfo;
        }
        // 保存账号信息到数据库
        emailInfo = this.saveEmailInfo(bindingVo);
        // 添加账号
        EmailEngineUtil.createAccount(bindingVo);
        // 查询邮箱文件夹
        List<EmailMailboxVo.MailboxesDTO> mailboxList = EmailEngineUtil.getMailboxList(email);
        // 添加邮箱文件夹
        List<EmailMailbox> emailMailboxList = this.saveEmailMailbox(mailboxList, emailInfo.getId(), email);
        // redis添加同步进度
        this.synchronizationProgressInitialization(email, bindingVo.getPages(), mailboxList);
        // 异步遍历文件夹下的所有邮件
        // this.asyncReadEmail(email, bindingVo.getPages(), emailMailboxList);
        return emailInfo;
    }

    @Transactional(rollbackFor = Exception.class)
    @Override
    public void listener(ListenerVo listenerVo) {

        log.info("监听到新消息：{}", JSON.toJSONString(listenerVo));

        switch (SendEventEnum.get(listenerVo.getEvent())) {
            case MESSAGE_NEW:
                this.handleMessageNewEvent(listenerVo);
                break;
            case MESSAGE_DELETED:
                this.handleMessageDeletedEvent(listenerVo);
                break;
            case MESSAGE_UPDATED:
                this.handleMessageUpdatedEvent(listenerVo);
                break;
            case MAILBOX_NEW:
                this.handleMailboxNewEvent(listenerVo);
                break;
            case MAILBOX_DELETED:
                this.handleMailboxDeleteEvent(listenerVo);
                break;
            case MAILBOX_RESET:
                this.handleMailboxResetEvent(listenerVo);
                break;
            default:
                log.error("监听到未知事件：{}", JSONObject.toJSONString(listenerVo));
        }

    }

    @Override
    public void submit(SubmitVo submitVo) {
        EmailEngineUtil.submit(submitVo);
    }

    /**
     * 添加邮件信息
     */
    private EmailInfo saveEmailInfo(BindingVo bindingVo) {
        EmailInfo emailInfo = BeanUtil.copyProperties(bindingVo, EmailInfo.class);
        emailInfoService.save(emailInfo);
        return emailInfo;
    }

    /**
     * 添加邮箱文件夹
     */
    private List<EmailMailbox> saveEmailMailbox(List<EmailMailboxVo.MailboxesDTO> mailboxDTOList, Long emailInfoId, String email) {

        List<EmailMailbox> mailboxList = mailboxDTOList.stream().map(item -> {
            EmailMailbox emailMailbox = BeanUtil.copyProperties(item, EmailMailbox.class);
            emailMailbox.setEmailInfoId(emailInfoId);
            emailMailbox.setEmail(email);
            emailMailbox.setIsShow(true);
            emailMailbox.setInitialize(false);
            return emailMailbox;
        }).distinct().collect(Collectors.toList());

        emailMailboxService.saveBatch(mailboxList);

        return mailboxList;
    }

    /**
     * 同步邮件进度初始化
     */
    private void synchronizationProgressInitialization(String email, int pages, List<EmailMailboxVo.MailboxesDTO> mailboxList) {

        ProgressVo progressVo = new ProgressVo();


        List<ProgressVo.Details> detailsList = BeanUtil.copyToList(mailboxList, ProgressVo.Details.class);

        int totalMessageCount = 0;

        for (ProgressVo.Details detail : detailsList) {
            detail.setMessages(detail.getMessages() > pages * 10 ? pages * 10 : detail.getMessages());
            detail.setCompleteMessageCount(0);
            detail.setPercentage(0);
            totalMessageCount += detail.getMessages();
        }

        progressVo.setEmail(email);
        progressVo.setTotalMessageCount(totalMessageCount);
        progressVo.setCompleteMessageCount(0);
        progressVo.setPercentage(0);
        progressVo.setDetailsList(detailsList);

        RedisCache.set(RedisConstant.PROGRESS_KEY + email, progressVo);
    }

    /**
     * 异步读取文件
     */
    private void asyncReadEmail(String email, int pages, List<EmailMailbox> emailMailboxList) {
        executor.execute(() -> {
            ThreadUtil.sleep(3000);

            // 记录上次消息id,防止循环查询分页过程中接收到了新邮件，出现邮件重复问题
            List<String> lastMessageIdList = new ArrayList<>();

            // 遍历每个邮箱
            for (EmailMailbox emailMailbox : emailMailboxList) {
                int page = 0;
                while (page < pages) {
                    MessageVo result = EmailEngineUtil.getMessageList(email, emailMailbox.getPath(), page);
                    List<MessageVo.MessagesDTO> messagesDTOList = result.getMessages();
                    if (messagesDTOList.size() > 0) {
                        List<EmailMessage> emailMessageList = this.saveBatchMessage(email, lastMessageIdList, emailMailbox, messagesDTOList);
                        this.synchronizationProgress(email, messagesDTOList, emailMailbox);

                        lastMessageIdList.clear();
                        for (EmailMessage emailMessage : emailMessageList) {
                            lastMessageIdList.add(emailMessage.getMessageId());
                        }
                    }
                    page++;
                }
            }
        });
    }

    /**
     * 批量保存邮件信息
     */
    private List<EmailMessage> saveBatchMessage(String email, List<String> lastMessageId, EmailMailbox emailMailbox, List<MessageVo.MessagesDTO> messageList) {
        List<EmailMessage> emailMessageList = new ArrayList<>();
        List<EmailMessageSend> emailMessageSendList = new ArrayList<>();
        List<EmailMessageAttachment> emailMessageAttachmentList = new ArrayList<>();

        for (MessageVo.MessagesDTO messagesDTO : messageList) {
            // 已添加过的邮件则跳过
            if (lastMessageId.contains(messagesDTO.getId())) {
                continue;
            }

            // 生成邮件实体
            EmailMessage emailMessage = this.createMessage(email, emailMailbox, messagesDTO);
            emailMessageList.add(emailMessage);

            // 添加推送信息
            this.addMessageSend(emailMessageSendList, messagesDTO, emailMessage.getId());

            // 下载并添加附件信息
            this.addMessageAttachment(email, emailMessageAttachmentList, messagesDTO, emailMessage.getId());
        }

        // 保存邮件
        emailMessageService.saveBatch(emailMessageList);
        // 保存推送人信息
        emailMessageSendService.saveBatch(emailMessageSendList);
        // 保存附件信息
        emailMessageAttachmentService.saveBatch(emailMessageAttachmentList);

        return emailMessageList;
    }

    /**
     * 生成邮件实体
     */
    private EmailMessage createMessage(String email, EmailMailbox emailMailbox, MessageVo.MessagesDTO messagesDTO) {

        EmailMessage emailMessage = new EmailMessage();
        emailMessage.setId(IdWorker.getId());
        emailMessage.setMessageId(messagesDTO.getId());
        emailMessage.setEmailInfoId(emailMailbox.getEmailInfoId());
        emailMessage.setEmail(emailMailbox.getEmail());
        emailMessage.setEmailMailboxId(emailMailbox.getId());
        emailMessage.setEmailMailboxPath(emailMailbox.getPath());
        emailMessage.setUnseen(messagesDTO.getUnseen());
        emailMessage.setFlagged(messagesDTO.getFlagged());
        emailMessage.setSubject(messagesDTO.getSubject());
        emailMessage.setFromDate(messagesDTO.getDate());

        // 发件人
        MessageVo.MessagesDTO.ToDTO from = messagesDTO.getFrom();
        if (ObjectUtils.isNotEmpty(from)) {
            emailMessage.setFromName(from.getName());
            emailMessage.setFromAddress(from.getAddress());
        }

        // 正文内容
        MessageDetailVo messageDetail = EmailEngineUtil.getMessageDetail(email, messagesDTO.getId());
        MessageDetailVo.TextDTO text = messageDetail.getText();
        if (ObjectUtils.isNotEmpty(text)) {
            emailMessage.setHtmlContent(text.getHtml());
        }

        return emailMessage;
    }

    /**
     * 添加推送信息
     */
    private void addMessageSend(List<EmailMessageSend> emailMessageSendList, MessageVo.MessagesDTO messagesDTO, Long emailMessageId) {
        // 收件人
        List<MessageVo.MessagesDTO.ToDTO> to = messagesDTO.getTo();
        if (ObjectUtils.isNotEmpty(to)) {
            for (MessageVo.MessagesDTO.ToDTO toDTO : to) {
                EmailMessageSend emailMessageSend = new EmailMessageSend();
                emailMessageSend.setEmailMessageId(emailMessageId);
                emailMessageSend.setMessageId(messagesDTO.getId());
                emailMessageSend.setType(SendConstant.TO);
                emailMessageSend.setName(toDTO.getName());
                emailMessageSend.setAddress(toDTO.getAddress());
                emailMessageSendList.add(emailMessageSend);
            }
        }

        // 抄送人
        List<MessageVo.MessagesDTO.ToDTO> cc = messagesDTO.getCc();
        if (ObjectUtils.isNotEmpty(cc)) {
            for (MessageVo.MessagesDTO.ToDTO toDTO : cc) {
                EmailMessageSend emailMessageSend = new EmailMessageSend();
                emailMessageSend.setEmailMessageId(emailMessageId);
                emailMessageSend.setMessageId(messagesDTO.getId());
                emailMessageSend.setType(SendConstant.CC);
                emailMessageSend.setName(toDTO.getName());
                emailMessageSend.setAddress(toDTO.getAddress());
                emailMessageSendList.add(emailMessageSend);
            }
        }

        // 密送人
        List<MessageVo.MessagesDTO.ToDTO> bcc = messagesDTO.getBcc();
        if (ObjectUtils.isNotEmpty(bcc)) {
            for (MessageVo.MessagesDTO.ToDTO toDTO : cc) {
                EmailMessageSend emailMessageSend = new EmailMessageSend();
                emailMessageSend.setEmailMessageId(emailMessageId);
                emailMessageSend.setMessageId(messagesDTO.getId());
                emailMessageSend.setType(SendConstant.BCC);
                emailMessageSend.setName(toDTO.getName());
                emailMessageSend.setAddress(toDTO.getAddress());
                emailMessageSendList.add(emailMessageSend);
            }
        }
    }

    /**
     * 下载并添加附件信息
     */
    private void addMessageAttachment(String email, List<EmailMessageAttachment> emailMessageAttachmentList,
                                      MessageVo.MessagesDTO messagesDTO, Long emailMessageId) {

        List<MessageVo.MessagesDTO.AttachmentsDTO> attachments = messagesDTO.getAttachments();
        if (ObjectUtils.isNotEmpty(attachments)) {
            for (MessageVo.MessagesDTO.AttachmentsDTO attachment : attachments) {
                String attachmentId = attachment.getId();

                String newFileName = attachmentId + "." + FileUtil.getSuffix(attachment.getFilename());

                // 下载附件
                try {
                    Boolean flag = RetryUtil.execute(() -> EmailEngineUtil.downloadAttachment(email, attachmentId, newFileName));

                    if (flag) {
                        EmailMessageAttachment emailMessageAttachment = new EmailMessageAttachment();
                        emailMessageAttachment.setEmailMessageId(emailMessageId);
                        emailMessageAttachment.setMessageId(messagesDTO.getId());
                        emailMessageAttachment.setName(attachment.getFilename());
                        emailMessageAttachment.setPath(email + "\\" + newFileName);
                        emailMessageAttachment.setAttachmentId(attachmentId);
                        emailMessageAttachment.setSize(attachment.getEncodedSize());
                        emailMessageAttachmentList.add(emailMessageAttachment);
                    }

                } catch (Exception e) {
                    log.error("下载附件失败：", e);
                }

            }
        }
    }

    /**
     * 同步邮件进度
     */
    private void synchronizationProgress(String email, List<MessageVo.MessagesDTO> messagesDTOList, EmailMailbox emailMailbox) {

        ProgressVo progressVo = RedisCache.get(RedisConstant.PROGRESS_KEY + email);
        int completeMessageCount = progressVo.getCompleteMessageCount() + messagesDTOList.size();
        progressVo.setCompleteMessageCount(completeMessageCount);
        progressVo.setPercentage(100 * completeMessageCount / progressVo.getTotalMessageCount());

        for (ProgressVo.Details details : progressVo.getDetailsList()) {
            if (details.getPath().equals(emailMailbox.getPath())) {
                int detailsCompleteMessageCount = details.getCompleteMessageCount() + messagesDTOList.size();
                details.setCompleteMessageCount(detailsCompleteMessageCount);
                if (details.getMessages() == 0) {
                    details.setPercentage(100);
                } else {
                    details.setPercentage(100 * detailsCompleteMessageCount / details.getMessages());
                }
                break;
            }
        }

        RedisCache.set(RedisConstant.PROGRESS_KEY + email, progressVo);
    }

    /**
     * 处理收到新邮件事件
     */
    private void handleMessageNewEvent(ListenerVo listenerVo) {
        // 查询邮箱
        String account = listenerVo.getAccount();
        String path = listenerVo.getPath();

        EmailMailbox emailMailbox;
        try {
            emailMailbox = RetryUtil.execute(() -> {
                EmailMailbox tempEmailMailbox = emailMailboxService.getOne(Wrappers.<EmailMailbox>lambdaQuery()
                        .eq(EmailMailbox::getPath, path)
                        .eq(EmailMailbox::getEmail, account)
                );
                Assert.notNull(tempEmailMailbox, "未找到邮箱信息");
                return tempEmailMailbox;
            });
        } catch (Exception e) {
            log.error("查找邮箱失败", e);
            return;
        }

        // 获取附件详情
        MessageVo.MessagesDTO messagesDTO = listenerVo.getData().toJavaObject(MessageVo.MessagesDTO.class);

        // 保存邮件
        EmailMessage emailMessage = this.createMessage(account, emailMailbox, messagesDTO);
        emailMessageService.save(emailMessage);

        // 保存推送人信息
        List<EmailMessageSend> emailMessageSendList = new ArrayList<>();
        this.addMessageSend(emailMessageSendList, messagesDTO, emailMessage.getId());
        emailMessageSendService.saveBatch(emailMessageSendList);

        // 保存附件信息
        List<EmailMessageAttachment> emailMessageAttachmentList = new ArrayList<>();
        this.addMessageAttachment(account, emailMessageAttachmentList, messagesDTO, emailMessage.getId());
        emailMessageAttachmentService.saveBatch(emailMessageAttachmentList);
    }

    /**
     * 处理删除邮件事件
     */
    private void handleMessageDeletedEvent(ListenerVo listenerVo) {
        MessageDeletedEventVo vo = listenerVo.getData().toJavaObject(MessageDeletedEventVo.class);
        emailMessageService.remove(
                Wrappers.<EmailMessage>lambdaQuery().eq(EmailMessage::getMessageId, vo.getId()));
    }

    /**
     * 邮件flag被改变事件
     */
    private void handleMessageUpdatedEvent(ListenerVo listenerVo) {
        log.error("邮件flag改变：{}", JSON.toJSONString(listenerVo));
    }

    /**
     * 新增文件夹事件
     */
    private void handleMailboxNewEvent(ListenerVo listenerVo) {

        EmailMailboxVo.MailboxesDTO mailboxesDTO = listenerVo.getData().toJavaObject(EmailMailboxVo.MailboxesDTO.class);

        EmailInfo emailInfo = emailInfoService.getOne(
                Wrappers.<EmailInfo>lambdaQuery().eq(EmailInfo::getEmail, listenerVo.getAccount()));

        EmailMailbox emailMailbox = new EmailMailbox();
        emailMailbox.setEmailInfoId(emailInfo.getId());
        emailMailbox.setEmail(emailInfo.getEmail());
        emailMailbox.setName(mailboxesDTO.getName());
        emailMailbox.setPath(mailboxesDTO.getPath());
        emailMailbox.setIsShow(true);
        emailMailbox.setInitialize(true);
        emailMailboxService.save(emailMailbox);
    }

    /**
     * 删除文件夹事件
     */
    private void handleMailboxDeleteEvent(ListenerVo listenerVo) {
        EmailMailboxVo.MailboxesDTO mailboxesDTO = listenerVo.getData().toJavaObject(EmailMailboxVo.MailboxesDTO.class);

        boolean remove = emailMailboxService.remove(Wrappers.<EmailMailbox>lambdaQuery()
                .eq(EmailMailbox::getPath, mailboxesDTO.getPath())
                .eq(EmailMailbox::getEmail, listenerVo.getAccount())
        );

        if (!remove) {
            log.error("文件夹删除失败：email '{}',path '{}'", listenerVo.getAccount(), mailboxesDTO.getPath());
        }

    }

    /**
     * 更新文件夹事件
     */
    public void handleMailboxResetEvent(ListenerVo listenerVo) {
        log.error("更新文件夹：{}", JSON.toJSONString(listenerVo));
    }

}
